/*
	descrypt.c
	Copyright (C) 2001 David Fifield

	This file is part of descrypt.

	descrypt is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	descrypt is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <config.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <stdarg.h>
#include <errno.h>
#include <getopt.h>
#include <sys/stat.h>

#include "descrypt.h"
#include "des.h"
#include "des_mode.h"

#if ENABLE_NLS
# include <libintl.h>
# include <locale.h>
# define _(s) gettext(s)
#else
# define _(s) s
# define textdomain(d)
# define bindtextdomain(d)
# define setlocale(d)
#endif

static char *prog_name = PACKAGE;
static unsigned char default_key[LONG_KEY_SIZE] = {
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01
};

int main(int argc, char *argv[])
{
	struct prog_option_t prog_option = {
		NULL,
		DES_ENCRYPT,
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		NULL,
		MODE_CBC,
		NULL,
	};
	int (*des_mode[])(FILE *out_file, FILE *in_file,
                      unsigned char key[LONG_KEY_SIZE],
                      unsigned int crypt_dir,
	                  unsigned char iv[BLOCK_SIZE]) = {
		des_mode_ecb,
		des_mode_cbc,
		des_mode_ofb,
		des_mode_cfb
	};
	FILE *in_file, *key_file, *out_file;
	unsigned char key[LONG_KEY_SIZE];
	unsigned char short_key[KEY_SIZE];
	struct stat in_stat, out_stat;
	int ret;

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);

	set_prog_option(argc, argv, &prog_option);

	if (prog_option.key_filename == NULL) {
		memcpy(key, default_key, LONG_KEY_SIZE);
	} else {
		key_file = fopen(prog_option.key_filename, "rb");
		if (key_file == NULL)
			error("%s: %s", prog_option.key_filename, strerror(errno));
		if (fread(key, 1, LONG_KEY_SIZE, key_file) != LONG_KEY_SIZE)
			error("%s: %s", prog_option.key_filename,
			      _("key file must be at least 8 bytes long"));
		fclose(key_file);
	}

	if (des_key_parity(key) == 0)
		warning("%s: %s", prog_option.key_filename,
		        _("key parity error"));

	if ((prog_option.in_filename != NULL)
	    && (prog_option.out_filename != NULL)) {
		if (stat(prog_option.in_filename, &in_stat) == -1)
			error("%s: %s", prog_option.in_filename, strerror(errno));
		if (stat(prog_option.out_filename, &out_stat) != -1) {
			if (in_stat.st_ino == out_stat.st_ino)
				error(_("`%s' and `%s' are the same file"),
			    	  prog_option.in_filename, prog_option.out_filename);
		}
	}

	if (prog_option.in_filename == NULL)
		in_file = stdin;
	else
		in_file = fopen(prog_option.in_filename, "rb");
	if (in_file == NULL)
		error("%s: %s", prog_option.in_filename, strerror(errno));

	if (prog_option.out_filename == NULL)
		out_file = stdout;
	else
		out_file = fopen(prog_option.out_filename, "wb");
	if (out_file == NULL)
		error("%s: %s", prog_option.out_filename, strerror(errno));

	ret = des_mode[prog_option.crypt_mode](out_file, in_file, key,
	                                       prog_option.crypt_dir,
	                                       prog_option.iv);
	if (ret != 0) {
		if (ret == 1)
			error("%s", strerror(errno));
		else
			error("%s: %s", prog_option.in_filename,
			      _("malformed ciphertext"));
	}

	fclose(in_file);
	fclose(out_file);

	return 0;
}

unsigned char *iv_parse(unsigned char iv[BLOCK_SIZE], const char *s)
{
	int i, length;
	char c;

	memset(iv, 0, BLOCK_SIZE);
	/* Skip leading zeros */
	while (*s == '0')
		s++;
	length = 0;
	while (*s != '\0') {
		s++;
		length++;
	}
	if (length > BLOCK_SIZE * 2)
		return NULL;
	i = BLOCK_SIZE - 1;
	for (;;) {
		if (length-- == 0)
			break;
		c = *--s;
		if (isdigit(c))
			c = c - '0';
		else if (isalpha(c))
			c = toupper(c) - 'A' + 10;
		else
			return NULL;
		if (c > 15)
			return NULL;
		iv[i] |= c;
		if (length-- == 0)
			break;
		c = *--s;
		if (isdigit(c))
			c = c - '0';
		else if (isalpha(c))
			c = toupper(c) - 'A' + 10;
		else
			return NULL;
		if (c > 15)
			return NULL;
		iv[i--] |= c << 4;
	}

	return iv;
}

void set_prog_option(int argc, char *argv[],
                     struct prog_option_t *prog_option)
{
	static struct option long_options[] = {
		{ "decrypt", no_argument, NULL, 'd' },
		{ "encrypt", no_argument, NULL, 'e' },
		{ "iv", required_argument, NULL, 'i' },
		{ "key", required_argument, NULL, 'k' },
		{ "mode", required_argument, NULL, 'm' },
		{ "output", required_argument, NULL, 'o' },
		{ "help", no_argument, NULL, 0 },
		{ "version", no_argument, NULL, 0 }
	};
	int option_index;
	char c;

	prog_name = argv[0];

	option_index = 0;
	while ((c = getopt_long(argc, argv, "dei:k:m:o:",
	                        long_options, &option_index)) != -1) {
		switch (c) {
			case 'd':	/* decrypt */
				prog_option->crypt_dir = DES_DECRYPT;
				break;
			case 'e':	/* encrypt */
				prog_option->crypt_dir = DES_ENCRYPT;
				break;
			case 'i':	/* iv */
				if (iv_parse(prog_option->iv, optarg) == NULL)
					error(_("invalid initialization vector: %s"), optarg);
				break;
			case 'k':	/* key */
				prog_option->key_filename = optarg;
				break;
			case 'm':	/* mode */
				if (strcmp(optarg, "ecb") == 0)
					prog_option->crypt_mode = MODE_ECB;
				else if (strcmp(optarg, "cbc") == 0)
					prog_option->crypt_mode = MODE_CBC;
				else if (strcmp(optarg, "ofb") == 0)
					prog_option->crypt_mode = MODE_OFB;
				else if (strcmp(optarg, "cfb") == 0)
					prog_option->crypt_mode = MODE_CFB;
				else
					error(_("bad mode: %s"), optarg);
				break;
			case 'o':	/* output */
				prog_option->out_filename = optarg;
				break;
			case 0:
				switch (option_index) {
					case 6:	/* help */
						print_help(stdout);
						exit(0);
						break;
					case 7:	/* version */
						print_version(stdout);
						exit(0);
						break;
					default:	/* ??? */
						error(_("\
Unrecognized long option: %s\n\
This can't happen."), long_options[option_index].name);
						break;
				}
				break;
			case '?':
				fprintf(stderr, _("\
Try `%s --help' for more information.\n"), PACKAGE);
				exit(1);
			default:
				error(_("this can't happen."));
				break;
		}
	}
	if (optind < argc) {
		if (strcmp(argv[optind], "-") != 0)
			prog_option->in_filename = argv[optind];
		optind++;
		if (optind < argc)
			error(_("you may specify only one input file"));
	}

	return;
}

void print_help(FILE *s)
{
	fprintf(s, _("\
Usage: %s [OPTION]... -k KEY [FILE]\n\
descrypt encrypts or decrypts FILE and prints the resulting ciphertext\n\
or plaintext to a file or to standard output. Encryption is the default.\n\
\n\
  -d, --decrypt       specify decryption\n\
  -e, --encrypt       specify encryption (default)\n\
  -i, --iv=IV         use the specified initialization vector. IV must be\n\
                        a string of 0 to 16 hexadecimal digits.\n\
  -k, --key=KEY       use the first 8 bytes of file KEY as the key.\n\
                        Without this option, use a key of all zeros.\n\
  -m, --mode=MODE     use the specified mode. MODE may be one of `ecb',\n\
                        `cbc' (default), `ofb' or `cfb'.\n\
  -o, --output=FILE   write output to FILE. Without this option, write to\n\
                        standard output.\n\
      --help          display this usage message and exit\n\
      --version       display version information and exit\n\
\n\
If no FILE is given or if FILE is `-', read standard input.\n\
\n\
Report bugs to <david@bamsoftware.com>.\n\
"), PACKAGE);

	return;
}

void print_version(FILE *s)
{
	fprintf(s, "%s %s\n", PACKAGE, VERSION);

	return;
}

void error(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "%s: ", prog_name);
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
	va_end(ap);
	exit(1);
}

void warning(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, _("%s: warning: "), prog_name);
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
	va_end(ap);

	return;
}

